"""Tests for the `tool` module."""

from abc import update_abstractmethods

import pytest
from pydantic import BaseModel

from mirascope.core.base._utils import DEFAULT_TOOL_DOCSTRING
from mirascope.core.base.tool import BaseTool


def test_base_tool() -> None:
    """Tests the `BaseTool` class."""

    class FormatBook(BaseTool):
        """Returns a formatted book title and author."""

        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    tool = FormatBook(title="The Name of the Wind", author="Patrick Rothfuss")
    assert tool._name() == "FormatBook"
    assert tool._description() == "Returns a formatted book title and author."
    assert tool.args == {"title": "The Name of the Wind", "author": "Patrick Rothfuss"}
    assert tool.call() == "The Name of the Wind by Patrick Rothfuss"


@pytest.mark.asyncio
async def test_base_tool_call_async() -> None:
    """Tests the `BaseTool.call` method with an async function."""

    class FormatBook(BaseTool):
        """Returns a formatted book title and author."""

        title: str
        author: str

        async def call(self) -> str:
            return f"{self.title} by {self.author}"

    tool = FormatBook(title="The Name of the Wind", author="Patrick Rothfuss")
    assert await tool.call() == "The Name of the Wind by Patrick Rothfuss"


def test_base_tool_no_doc() -> None:
    """Tests the `BaseTool` class with no docstring."""
    from mirascope.core.base import BaseTool

    class FormatBook(BaseTool):
        title: str

    assert FormatBook._description() == DEFAULT_TOOL_DOCSTRING


def test_base_tool_custom_name() -> None:
    """Tests the `BaseTool` class with a custom name."""
    from mirascope.core.base import BaseTool

    class FormatBook(BaseTool):
        __custom_name__ = "format_book"

    assert FormatBook._name() == "format_book"


def test_base_tool_model_tool_schema() -> None:
    """Tests the `BaseTool.model_tool_schema` class method."""

    class Book(BaseModel):
        title: str
        author: str

    class FormatBook(BaseTool):
        book: Book
        genre: str

    assert FormatBook.model_tool_schema() == {
        "$defs": {
            "Book": {
                "properties": {
                    "title": {"type": "string"},
                    "author": {"type": "string"},
                },
                "required": ["title", "author"],
                "type": "object",
            }
        },
        "properties": {"book": {"$ref": "#/$defs/Book"}, "genre": {"type": "string"}},
        "required": ["book", "genre"],
        "type": "object",
    }


def test_base_tool_type_conversion() -> None:
    """Tests the `BaseTool.type_from...` class methods."""

    def format_book(title: str, author: str) -> str:
        return f"{title} by {author}"

    tool_type = BaseTool.type_from_fn(format_book)
    assert tool_type.__name__ == "format_book"
    tool = tool_type(title="The Name of the Wind", author="Patrick Rothfuss")  # type: ignore
    assert tool.call() == "The Name of the Wind by Patrick Rothfuss"

    class FormatBook(BaseModel):
        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    tool_type = BaseTool.type_from_base_model_type(FormatBook)
    assert tool_type.__name__ == "FormatBook"
    tool = tool_type(title="The Name of the Wind", author="Patrick Rothfuss")  # type: ignore
    assert isinstance(tool, BaseTool)
    assert tool.call() == "The Name of the Wind by Patrick Rothfuss"

    tool_type = BaseTool.type_from_base_type(str)
    tool_type.call = lambda self: self.value  # pyright: ignore [reportAttributeAccessIssue]
    update_abstractmethods(tool_type)
    assert tool_type.__name__ == "str"
    tool = tool_type(value="The Name of the Wind")  # type: ignore
    assert tool.call() == "The Name of the Wind"
    assert isinstance(tool, BaseTool)



from mirascope.core import BasePrompt, prompt_template


@prompt_template(
    """
    SYSTEM: You are the world's greatest librarian
    USER: Recommend a {genre} book
    """
)
class RecommendBookPrompt(BasePrompt):
    genre: str


prompt = RecommendBookPrompt(genre="fantasy")
print(prompt)
# > SYSTEM: You are the world's greatest librarian
#   USER: Recommend a fantasy book
print(prompt.message_params())
# > [BaseMessageParam(role='system', content="You are the world's greatest librarian"), BaseMessageParam(role='user', content='Recommend a fantasy book')]



from unittest.mock import MagicMock, patch

import pytest
from pydantic import BaseModel

from mirascope.core.base.call_response import BaseCallResponse
from mirascope.core.base.metadata import Metadata
from mirascope.core.base.stream import BaseStream
from mirascope.core.base.structured_stream import BaseStructuredStream
from mirascope.integrations.langfuse import _utils
from mirascope.integrations.langfuse._utils import ModelUsage


def test_get_call_response_observation() -> None:
    """Tests the `get_call_response_observation` function."""
    mock_result = MagicMock(spec=BaseCallResponse)
    mock_result.model = "test_model"
    mock_result.prompt_template = "test_prompt_template"
    mock_result.response = "test_response"
    mock_result.message_param = {"role": "assistant", "content": "test_content"}
    mock_fn = MagicMock(__name__="mock_fn")
    mock_fn._metadata = Metadata(tags={"tag1"})
    call_response_observation = _utils.get_call_response_observation(
        mock_result, mock_fn
    )
    assert call_response_observation["name"] == "mock_fn with test_model"
    assert call_response_observation["input"] == "test_prompt_template"
    assert call_response_observation["metadata"] == "test_response"
    assert call_response_observation["tags"] == {"tag1"}
    assert call_response_observation["model"] == "test_model"
    assert call_response_observation["output"] == "test_content"


@patch(
    "mirascope.integrations.langfuse._utils.langfuse_context.update_current_observation",
    new_callable=MagicMock,
)
@patch(
    "mirascope.integrations.langfuse._utils.get_call_response_observation",
    new_callable=MagicMock,
)
def test_handle_call_response(
    mock_get_call_response_observation: MagicMock,
    mock_update_current_observation: MagicMock,
) -> None:
    """Tests the `handle_call_response` function."""
    mock_fn = MagicMock(__name__="mock_fn")

    result = MagicMock(spec=BaseCallResponse)
    _utils.handle_call_response(result, mock_fn, None)
    mock_get_call_response_observation.assert_called_once_with(result, mock_fn)
    mock_update_current_observation.assert_called_once_with(
        **mock_get_call_response_observation.return_value,
        usage=ModelUsage(
            input=result.input_tokens,
            output=result.output_tokens,
            unit="TOKENS",
        ),
    )


@patch(
    "mirascope.integrations.langfuse._utils.langfuse_context.update_current_observation",
    new_callable=MagicMock,
)
@patch(
    "mirascope.integrations.langfuse._utils.get_call_response_observation",
    new_callable=MagicMock,
)
@pytest.mark.asyncio
async def test_handle_call_response_async(
    mock_get_call_response_observation: MagicMock,
    mock_update_current_observation: MagicMock,
) -> None:
    """Tests the `handle_call_response` function."""
    mock_fn = MagicMock(__name__="mock_fn")

    result = MagicMock(spec=BaseCallResponse)
    await _utils.handle_call_response_async(result, mock_fn, None)
    mock_get_call_response_observation.assert_called_once_with(result, mock_fn)
    mock_update_current_observation.assert_called_once_with(
        **mock_get_call_response_observation.return_value,
        usage=ModelUsage(
            input=result.input_tokens,
            output=result.output_tokens,
            unit="TOKENS",
        ),
    )


@patch(
    "mirascope.integrations.langfuse._utils.langfuse_context.update_current_observation",
    new_callable=MagicMock,
)
@patch(
    "mirascope.integrations.langfuse._utils.get_call_response_observation",
    new_callable=MagicMock,
)
def test_handle_stream(
    mock_get_call_response_observation: MagicMock,
    mock_update_current_observation: MagicMock,
) -> None:
    """Tests the `handle_stream` function."""
    mock_fn = MagicMock(__name__="mock_fn")

    result = MagicMock(spec=BaseStream)
    mock_construct_call_response = MagicMock()
    mock_construct_call_response_return_value = {}
    mock_construct_call_response.return_value = (
        mock_construct_call_response_return_value
    )
    result.construct_call_response = mock_construct_call_response
    _utils.handle_stream(result, mock_fn, None)
    mock_get_call_response_observation.assert_called_once_with(
        mock_construct_call_response(), mock_fn
    )
    mock_update_current_observation.assert_called_once_with(
        **mock_construct_call_response_return_value,
        usage=ModelUsage(
            input=result.input_tokens,
            output=result.output_tokens,
            unit="TOKENS",
        ),
    )


@patch(
    "mirascope.integrations.langfuse._utils.handle_stream",
    new_callable=MagicMock,
)
@pytest.mark.asyncio
async def test_handle_stream_async(
    mock_handle_stream: MagicMock,
) -> None:
    """Tests the `handle_stream_async` function."""
    mock_fn = MagicMock(__name__="mock_fn")

    result = MagicMock(spec=BaseStream)
    await _utils.handle_stream_async(result, mock_fn, None)
    mock_handle_stream.assert_called_once_with(result, mock_fn, None)


@patch(
    "mirascope.integrations.langfuse._utils.langfuse_context.update_current_observation",
    new_callable=MagicMock,
)
@patch(
    "mirascope.integrations.langfuse._utils.get_call_response_observation",
    new_callable=MagicMock,
)
def test_handle_response_model(
    mock_get_call_response_observation: MagicMock,
    mock_update_current_observation: MagicMock,
) -> None:
    """Tests the `handle_response_model` function."""
    mock_fn = MagicMock(__name__="mock_fn")
    result = MagicMock(spec=BaseModel)
    response = MagicMock(spec=BaseCallResponse)
    result._response = response

    _utils.handle_response_model(result, mock_fn, None)
    mock_get_call_response_observation.assert_called_once_with(response, mock_fn)
    mock_update_current_observation.assert_called_once_with(
        **mock_get_call_response_observation.return_value,
        usage=ModelUsage(
            input=response.input_tokens,
            output=response.output_tokens,
            unit="TOKENS",
        ),
        output=result,
    )


@patch(
    "mirascope.integrations.langfuse._utils.langfuse_context.update_current_observation",
    new_callable=MagicMock,
)
def test_handle_response_model_base_type(
    mock_update_current_observation: MagicMock,
) -> None:
    """Tests the `handle_response_model` function with `BaseType` result."""
    mock_fn = MagicMock(__name__="mock_fn")
    result = MagicMock(spec=str)

    _utils.handle_response_model(result, mock_fn, None)
    mock_update_current_observation.assert_called_once_with(
        output=result,
    )


@patch(
    "mirascope.integrations.langfuse._utils.langfuse_context.update_current_observation",
    new_callable=MagicMock,
)
@patch(
    "mirascope.integrations.langfuse._utils.get_call_response_observation",
    new_callable=MagicMock,
)
def test_handle_structured_stream(
    mock_get_call_response_observation: MagicMock,
    mock_update_current_observation: MagicMock,
) -> None:
    """Tests the `handle_structured_stream` function."""
    mock_fn = MagicMock(__name__="mock_fn")
    result = MagicMock(spec=BaseStructuredStream)
    result.constructed_response_model = MagicMock()
    stream = MagicMock(spec=BaseStream)
    result.stream = stream
    mock_construct_call_response = MagicMock()
    mock_construct_call_response_return_value = {}
    mock_construct_call_response.return_value = (
        mock_construct_call_response_return_value
    )
    stream.construct_call_response = mock_construct_call_response
    _utils.handle_structured_stream(result, mock_fn, None)
    mock_get_call_response_observation.assert_called_once_with(
        mock_construct_call_response(), mock_fn
    )
    mock_update_current_observation.assert_called_once_with(
        **mock_construct_call_response_return_value,
        usage=ModelUsage(
            input=stream.input_tokens,
            output=stream.output_tokens,
            unit="TOKENS",
        ),
        output=result.constructed_response_model,
    )


@patch(
    "mirascope.integrations.langfuse._utils.langfuse_context.update_current_observation",
    new_callable=MagicMock,
)
@patch(
    "mirascope.integrations.langfuse._utils.get_call_response_observation",
    new_callable=MagicMock,
)
@pytest.mark.asyncio
async def test_handle_response_model_async(
    mock_get_call_response_observation: MagicMock,
    mock_update_current_observation: MagicMock,
) -> None:
    """Tests the `handle_response_model` function."""
    mock_fn = MagicMock(__name__="mock_fn")
    result = MagicMock(spec=BaseModel)
    response = MagicMock(spec=BaseCallResponse)
    result._response = response

    await _utils.handle_response_model_async(result, mock_fn, None)
    mock_get_call_response_observation.assert_called_once_with(response, mock_fn)
    mock_update_current_observation.assert_called_once_with(
        **mock_get_call_response_observation.return_value,
        usage=ModelUsage(
            input=response.input_tokens,
            output=response.output_tokens,
            unit="TOKENS",
        ),
        output=result,
    )


@patch(
    "mirascope.integrations.langfuse._utils.handle_structured_stream",
    new_callable=MagicMock,
)
@pytest.mark.asyncio
async def test_handle_structured_stream_async(
    mock_handle_structured_stream: MagicMock,
) -> None:
    """Tests the `handle_structured_stream_async` function."""
    mock_fn = MagicMock(__name__="mock_fn")
    result = MagicMock(spec=BaseStructuredStream)

    await _utils.handle_structured_stream_async(result, mock_fn, None)
    mock_handle_structured_stream.assert_called_once_with(result, mock_fn, None)



"""Tests the `openai._utils.get_json_output` module."""

import pytest
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types.chat.chat_completion import Choice
from openai.types.chat.chat_completion_chunk import (
    Choice as ChunkChoice,
)
from openai.types.chat.chat_completion_chunk import (
    ChoiceDelta,
    ChoiceDeltaToolCall,
    ChoiceDeltaToolCallFunction,
)
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)

from mirascope.core.openai._utils._get_json_output import get_json_output
from mirascope.core.openai.call_response import OpenAICallResponse
from mirascope.core.openai.call_response_chunk import OpenAICallResponseChunk


def test_get_json_output_call_response() -> None:
    """Tests the `get_json_output` function with a call response."""
    tool_call = ChatCompletionMessageToolCall(
        id="id",
        function=Function(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
        type="function",
    )
    choices = [
        Choice(
            finish_reason="stop",
            index=0,
            message=ChatCompletionMessage(
                content="json_output", role="assistant", tool_calls=[tool_call]
            ),
        )
    ]
    completion = ChatCompletion(
        id="id",
        choices=choices,
        created=0,
        model="gpt-4o",
        object="chat.completion",
    )
    call_response = OpenAICallResponse(
        metadata={},
        response=completion,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert get_json_output(call_response, json_mode=True) == "json_output"
    assert (
        get_json_output(call_response, json_mode=False)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )

    completion.choices[0].message.content = None
    completion.choices[0].message.tool_calls = None
    with pytest.raises(
        ValueError, match="No tool call or JSON object found in response."
    ):
        get_json_output(call_response, json_mode=False)


def test_get_json_output_call_response_chunk() -> None:
    """Tests the `get_json_output` function with a call response chunk."""
    tool_call = ChoiceDeltaToolCall(
        index=0,
        id="id",
        function=ChoiceDeltaToolCallFunction(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="function",
        ),
        type="function",
    )
    choices = [
        ChunkChoice(
            delta=ChoiceDelta(content="json_output", tool_calls=[tool_call]),
            index=0,
            finish_reason="stop",
        )
    ]
    chunk = ChatCompletionChunk(
        id="id",
        choices=choices,
        created=0,
        model="gpt-4o",
        object="chat.completion.chunk",
    )
    call_response_chunk = OpenAICallResponseChunk(chunk=chunk)
    assert get_json_output(call_response_chunk, json_mode=True) == "json_output"
    assert (
        get_json_output(call_response_chunk, json_mode=False)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )

    chunk.choices[0].delta.tool_calls = None
    assert get_json_output(call_response_chunk, json_mode=False) == ""



import sqlite3

from dotenv import load_dotenv
from openai.types.chat import ChatCompletionMessageParam
from pydantic import BaseModel, SkipValidation

from mirascope.core import openai, prompt_template

con = sqlite3.connect("database.db")
cur = con.cursor()
# ONE TIME SETUP
cur.execute(
    "CREATE TABLE movie(id INTEGER AUTO_INCREMENT PRIMARY KEY, title, year, score)"
)
cur.execute(
    "CREATE TABLE actor(id INTEGER AUTO_INCREMENT PRIMARY KEY, movie_id, name, age)"
)
cur.execute("""
    INSERT INTO movie (title, year, score) VALUES
        ('Monty Python and the Holy Grail', 1975, 8.2),
        ('And Now for Something Completely Different', 1971, 7.5)
""")
con.commit()
# END OF ONE TIME SETUP
load_dotenv()


def run_query(query: str):
    """A SELECT query to run."""
    print(query)
    try:
        cur.execute(query)
        res = cur.fetchall()
        return res
    except sqlite3.Error as e:
        return str(e)


def execute_query(query: str):
    """An INSERT, UPDATE, or DELETE query to execute."""
    print(query)
    try:
        cur.execute(query)
        con.commit()
        return "Query executed successfully."
    except sqlite3.Error as e:
        return str(e)


class DatabaseAdministrator(BaseModel):
    messages: SkipValidation[list[ChatCompletionMessageParam]]

    @openai.call(model="gpt-4o-mini", tools=[run_query, execute_query])
    @prompt_template(
        """
        SYSTEM:
        You are a database admin that is an expert at writing SQLite queries.

        Your schema is as follows:
        TABLE movie(id, title, year, score)
        TABLE actor(id, movie_id, name, age)

        Translate the following text into a SQLite query. If the table or columns do not exist,
        return None. Then execute the query and return the results.

        Example:
        text: Show me the titles of all movies.
        query: SELECT title FROM movie;

        Example:
        text: Add Gone with the Wind to the movie table.
        query: INSERT INTO movie (title) VALUES ('Gone with the Wind', 1939, 8.1);

        MESSAGES: {self.messages}

        USER:
        {text}
        """
    )
    def _step(self, text: str): ...

    def _get_response(self, question: str = ""):
        response = self._step(question)
        tools_and_outputs = []
        if tools := response.tools:
            for tool in tools:
                output = tool.call()
                tools_and_outputs.append((tool, str(output)))
        else:
            print("(Assistant):", response.content)
            return
        if response.user_message_param:
            self.messages.append(response.user_message_param)
        self.messages += [
            response.message_param,
            *response.tool_message_params(tools_and_outputs),
        ]
        return self._get_response("")

    def run(self):
        while True:
            question = input("(User): ")
            if question == "exit":
                break
            self._get_response(question)


DatabaseAdministrator(messages=[]).run()



"""Tests the `_utils.convert_function_to_base_tool` module."""

import pytest

from mirascope.core.base._utils._convert_function_to_base_tool import (
    convert_function_to_base_tool,
)
from mirascope.core.base.tool import BaseTool


def test_convert_function_to_base_tool() -> None:
    """Tests the `convert_function_to_base_tool` function."""

    def format_book(title: str, author: str) -> str:
        """Returns the title and author nicely formatted."""
        return f"{title} by {author}"

    tool_type = convert_function_to_base_tool(format_book, BaseTool)
    assert tool_type._name() == "format_book"
    assert tool_type._description() == "Returns the title and author nicely formatted."
    assert "title" in tool_type.model_fields
    assert "author" in tool_type.model_fields
    tool = tool_type(title="The Name of the Wind", author="Patrick Rothfuss")  # type: ignore
    assert tool.call() == "The Name of the Wind by Patrick Rothfuss"


@pytest.mark.asyncio
async def test_convert_async_function_to_base_tool() -> None:
    """Tests the `convert_function_to_base_tool` function with async functions."""

    async def format_book(title: str, author: str) -> str:
        """Returns the title and author nicely formatted."""
        return f"{title} by {author}"

    tool_type = convert_function_to_base_tool(format_book, BaseTool)
    tool = tool_type(title="The Name of the Wind", author="Patrick Rothfuss")  # type: ignore
    assert await tool.call() == "The Name of the Wind by Patrick Rothfuss"


def test_convert_class_function_to_base_tool_fully_documented() -> None:
    """Tests `convert_function_to_base_tool` with a fully documented class function."""

    def format_book(title: str = "default_title") -> None:
        """Returns the title and author nicely formatted.

        Examples:
            {"title": "The Name of the Wind"}

        Args:
            title: The title of the book.
        """

    tool_type = convert_function_to_base_tool(format_book, BaseTool)
    assert "title" in tool_type.model_fields
    title_field = tool_type.model_fields["title"]
    assert title_field.default == "default_title"
    assert title_field.description == "The title of the book."
    assert "json_schema_extra" in tool_type.model_config
    assert tool_type.model_config["json_schema_extra"]["examples"] == [  # type: ignore
        {"title": "The Name of the Wind"}
    ]


def test_convert_function_to_base_tool_with_protected_namespace_field() -> None:
    """Tests `convert_function_to_base_tool` with a protected namespace field."""

    def format_book(model_title: str) -> None:
        """Tool with protected name model_*"""

    tool_type = convert_function_to_base_tool(format_book, BaseTool)
    assert "aliased_model_title" in tool_type.model_fields


def test_convert_function_to_base_tool_with_self_argument() -> None:
    """Tests `convert_function_to_base_tool` with a function that takes `self`."""

    def format_book(self):
        return self.title

    tool_type = convert_function_to_base_tool(format_book, BaseTool)
    tool_type.title = "The Name of the Wind"  # pyright: ignore [reportAttributeAccessIssue]
    tool = tool_type()
    assert tool.call() == "The Name of the Wind"


def test_convert_function_to_base_tool_with_cls_argument() -> None:
    """Tests `convert_function_to_base_tool` with a function that takes `cls`."""

    def format_book(cls=None) -> str:
        return "The Name of the Wind"

    tool_type = convert_function_to_base_tool(format_book, BaseTool)
    tool = tool_type()
    assert tool.call() == "The Name of the Wind"


def test_convert_function_to_base_tool_missing_type_annotation() -> None:
    """Tests `convert_function_to_base_tool` with a missing a type annotation."""

    def format_book(title) -> None:
        """Tool with missing type annotation."""

    with pytest.raises(ValueError):
        convert_function_to_base_tool(format_book, BaseTool)


def test_convert_function_to_base_tool_incorrect_parameter_name() -> None:
    """Tests `convert_function_to_base_tool` with an incorrect parameter name."""

    def format_book(title: str) -> None:
        """Returns the title and author nicely formatted.

        Args:
            titles: The title of the book.
        """

    with pytest.raises(ValueError):
        convert_function_to_base_tool(format_book, BaseTool)


def test_convert_function_to_base_tool_missing_parameter_description() -> None:
    """Tests `convert_function_to_base_tool` with a missing parameter description."""

    def format_book(title: str) -> None:
        """Returns the title and author nicely formatted.

        Args:
            title:
        """

    with pytest.raises(ValueError):
        convert_function_to_base_tool(format_book, BaseTool)



from pydantic import BaseModel

from mirascope.core import openai, prompt_template
from mirascope.integrations.langfuse import with_langfuse


class Book(BaseModel):
    title: str
    author: str


@with_langfuse()
@openai.call(model="gpt-4o-mini", response_model=Book)
@prompt_template("Recommend a {genre} book.")
def recommend_book(genre: str): ...


print(recommend_book("fantasy"))



from mirascope.core import openai
from mirascope.core.base.prompt import prompt_template

few_shot_examples = [
    {
        "original_question": "Which position did Knox Cunningham hold from May 1955 to Apr 1956?",
        "stepback_question": "Which positions have Knox Cunningham held in his career?",
    },
    {
        "original_question": "Who was the spouse of Anna Karina from 1968 to 1974?",
        "stepback_question": "Who were the spouses of Anna Karina?",
    },
    {
        "original_question": "Which team did Thierry Audel play for from 2007 to 2008?",
        "stepback_question": "Which teams did Thierry Audel play for in his career?",
    },
    {
        "original_question": "What was the operator of GCR Class 11E from 1913 to Dec 1922?",
        "stepback_question": "What were the operators of GCR Class 11E in history?",
    },
    {
        "original_question": "Which country did Sokolovsko belong to from 1392 to 1525?",
        "stepback_question": "Which countries did Sokolovsko belong to in history?",
    },
    {
        "original_question": "when was the last time a team from canada won the stanley cup as of 2002",
        "stepback_question": "which years did a team from canada won the stanley cup as of 2002",
    },
    {
        "original_question": "when did england last get to the semi final in a world cup as of 2019",
        "stepback_question": "which years did england get to the semi final in a world cup as of 2019?",
    },
    {
        "original_question": "what is the biggest hotel in las vegas nv as of November 28, 1993",
        "stepback_question": "what is the size of the hotels in las vegas nv as of November 28, 1993",
    },
    {
        "original_question": "who has scored most runs in t20 matches as of 2017",
        "stepback_question": "What are the runs of players in t20 matches as of 2017",
    },
]

stepback_prompt = """You are an expert at world knowledge. Your task is to step
back and paraphrase a question to a more generic step-back question, which is
easier to answer. Here are a few examples:"""


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    SYSTEM: {stepback_prompt_and_examples}
    USER: {query}
    """
)
def get_stepback_question(
    query: str, num_examples: int = 0
) -> openai.OpenAIDynamicConfig:
    """Gets the generic, step-back version of a query."""
    if num_examples < 0 or num_examples > len(few_shot_examples):
        raise ValueError(
            "num_examples cannot be negative or greater than number of available examples."
        )
    example_prompts = ""
    for i in range(num_examples):
        example_prompts += (
            f"Original Question: {few_shot_examples[i]['original_question']}\n"
        )
        example_prompts += (
            f"Stepback Question: {few_shot_examples[i]['stepback_question']}\n"
        )
    return {
        "computed_fields": {
            "stepback_prompt_and_examples": f"{stepback_prompt}\n{example_prompts}"
            if num_examples
            else ""
        }
    }


@openai.call(model="gpt-4o-mini")
@prompt_template("""{query}""")
def call(query: str):
    """A standard call to OpenAI."""


@openai.call(model="gpt-4o-mini")
@prompt_template(
    """
    You are an expert of world knowledge. I am going to ask you a question.
    Your response should be comprehensive and not contradicted with the
    following context if they are relevant. Otherwise, ignore them if they are
    not relevant.

    {stepback_response}

    Original Question: {query}
    Answer:
    """
)
def stepback(query: str, num_examples: int) -> openai.OpenAIDynamicConfig:
    """Executes the flow of the Step-Back technique."""
    stepback_question = get_stepback_question(
        query=query, num_examples=num_examples
    ).content
    stepback_response = call(query=stepback_question).content
    # Uncomment to see intermediate responses
    # print(stepback_question)
    # print(stepback_response)
    return {"computed_fields": {"stepback_response": stepback_response}}


query = """Who is the highest paid player in the nba this season as of 2017"""

print(stepback(query=query, num_examples=len(few_shot_examples)))
# > The highest-paid player in the NBA during the 2017 season was Stephen Curry of the Golden State Warriors, earning approximately $34.7 million.


# Intermediate Responses

# stepback_question = get_stepback_question(query=query, num_examples=num_examples)
# > Who were the highest paid players in the NBA during the 2017 season?

# stepback_response = call(stepback_question)
# > During the 2017 NBA season, the highest-paid players primarily included superstars with maximum contracts. Here are some of the top earners for that season:
#
#   1. **Stephen Curry** - Golden State Warriors: Approximately $34.7 million
#   2. **LeBron James** - Cleveland Cavaliers: Approximately $31.3 million
#   3. **Kevin Durant** - Golden State Warriors: Approximately $25.0 million
#   4. **James Harden** - Houston Rockets: Approximately $28.3 million
#   5. **Russell Westbrook** - Oklahoma City Thunder: Approximately $26.5 million



import os

from mirascope.core import BasePrompt, openai

# Enable docstring prompt templates
os.environ["MIRASCOPE_DOCSTRING_PROMPT_TEMPLATE"] = "ENABLED"


class BookRecommendationPrompt(BasePrompt):
    """Recommend a {genre} book"""

    genre: str


prompt = BookRecommendationPrompt(genre="fantasy")
print(prompt)
# > Recommend a fantasy book


@openai.call(model="gpt-4o-mini")
def recommend_book(genre: str):
    """Recommend a {genre} book"""
    ...


response = recommend_book("mystery")
print(response.content)
# > Here's a recommendation for a fantasy book: ...



from mirascope.core import anthropic, prompt_template
from mirascope.integrations.langfuse import with_langfuse


def format_book(title: str, author: str):
    return f"{title} by {author}"


@with_langfuse()
@anthropic.call(model="claude-3-5-sonnet-20240620", tools=[format_book])
@prompt_template("Recommend a {genre} book.")
def recommend_book(genre: str): ...


print(recommend_book("fantasy"))



import asyncio

from mirascope.core import BasePrompt, openai, prompt_template


@prompt_template("Analyze the sentiment of the following text: {text}")
class SentimentAnalysisPrompt(BasePrompt):
    text: str


async def main():
    prompt = SentimentAnalysisPrompt(text="I love using Mirascope!")
    result = await prompt.run_async(openai.call(model="gpt-4o-mini"))
    print(result.content)


asyncio.run(main())



import re


def regex_eval(output: str, pattern: str) -> bool:
    return bool(re.search(pattern, output))


# Example usage
output = "My email is john.doe@example.com"
email_pattern = r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b"
result = regex_eval(output, email_pattern)
print(result)  # Output: True



"""Tests the `gemini.call_response` module."""

from google.ai.generativelanguage import (
    Candidate,
    Content,
    FunctionCall,
    GenerateContentResponse,
    Part,
)
from google.generativeai.protos import FunctionResponse  # type: ignore
from google.generativeai.types import (  # type: ignore
    GenerateContentResponse as GenerateContentResponseType,
)

from mirascope.core.gemini.call_response import GeminiCallResponse
from mirascope.core.gemini.tool import GeminiTool


def test_gemini_call_response() -> None:
    """Tests the `GeminiCallResponse` class."""
    response = GenerateContentResponseType.from_response(
        GenerateContentResponse(
            candidates=[
                Candidate(
                    finish_reason=1,
                    content=Content(
                        parts=[Part(text="The author is Patrick Rothfuss")],
                        role="model",
                    ),
                )
            ]
        )
    )
    call_response = GeminiCallResponse(
        metadata={},
        response=response,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    call_response._model = "gemini-flash-1.5"
    assert call_response._provider == "gemini"
    assert call_response.content == "The author is Patrick Rothfuss"
    assert call_response.finish_reasons == ["STOP"]
    assert call_response.model == "gemini-flash-1.5"
    assert call_response.id is None
    assert call_response.tools is None
    assert call_response.tool is None
    assert call_response.usage is None
    assert call_response.input_tokens is None
    assert call_response.output_tokens is None
    assert call_response.cost is None
    assert call_response.message_param == {
        "role": "model",
        "parts": [Part(text="The author is Patrick Rothfuss")],
    }


def test_gemini_call_response_with_tools() -> None:
    """Tests the `GeminiCallResponse` class with tools."""

    class FormatBook(GeminiTool):
        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    response = GenerateContentResponseType.from_response(
        GenerateContentResponse(
            candidates=[
                Candidate(
                    finish_reason=1,
                    content=Content(
                        parts=[
                            Part(
                                function_call=FunctionCall(
                                    name="FormatBook",
                                    args={
                                        "title": "The Name of the Wind",
                                        "author": "Patrick Rothfuss",
                                    },
                                )
                            ),
                        ],
                        role="model",
                    ),
                )
            ]
        )
    )
    call_response = GeminiCallResponse(
        metadata={},
        response=response,
        tool_types=[FormatBook],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    tools = call_response.tools
    tool = call_response.tool
    assert tools and len(tools) == 1 and tools[0] == tool
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    output = tool.call()
    assert output == "The Name of the Wind by Patrick Rothfuss"
    assert call_response.tool_message_params([(tool, output)]) == [
        FunctionResponse(name="FormatBook", response={"result": output})
    ]



import json

from mirascope.core import openai, prompt_template


@openai.call(model="gpt-4o-mini", json_mode=True)
@prompt_template(
    """
    Provide the following information about {book_title}:
    - author
    - date published
    - genre
    """
)
def get_book_info(book_title: str): ...


response = get_book_info("The Great Gatsby")
book_info = json.loads(response.content)
print(book_info["author"])



from mirascope.core import BasePrompt, prompt_template


@prompt_template(
    """
    SYSTEM: You are the world's greatest librarian.
    USER: {query}
    """
)
class BookRecommendationPrompt(BasePrompt):
    query: str | None


prompt = BookRecommendationPrompt(query=None)
print(prompt.message_params())
# >[BaseMessageParam(role='system', content="You are the world's greatest librarian.")]



from mirascope.core import openai, prompt_template


@openai.call("gpt-4o-mini")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str) -> openai.OpenAIDynamicConfig:
    return {"metadata": {"tags": {f"genre:{genre}", "type:recommendation"}}}


response = recommend_book("horror")
print(response.metadata)



"""Tests the `openai.call_response` module."""

from openai.types.chat import ChatCompletion, ChatCompletionToolMessageParam
from openai.types.chat.chat_completion import Choice
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)
from openai.types.completion_usage import CompletionUsage

from mirascope.core.openai.call_response import OpenAICallResponse
from mirascope.core.openai.tool import OpenAITool


def test_openai_call_response() -> None:
    """Tests the `OpenAICallResponse` class."""
    choices = [
        Choice(
            finish_reason="stop",
            index=0,
            message=ChatCompletionMessage(content="content", role="assistant"),
        )
    ]
    usage = CompletionUsage(completion_tokens=1, prompt_tokens=1, total_tokens=2)
    completion = ChatCompletion(
        id="id",
        choices=choices,
        created=0,
        model="gpt-4o",
        object="chat.completion",
        usage=usage,
    )
    call_response = OpenAICallResponse(
        metadata={},
        response=completion,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert call_response._provider == "openai"
    assert call_response.content == "content"
    assert call_response.finish_reasons == ["stop"]
    assert call_response.model == "gpt-4o"
    assert call_response.id == "id"
    assert call_response.usage == usage
    assert call_response.input_tokens == 1
    assert call_response.output_tokens == 1
    assert call_response.cost == 2e-5
    assert call_response.message_param == {
        "content": "content",
        "role": "assistant",
        "tool_calls": None,
        "refusal": None,
    }
    assert call_response.tools is None
    assert call_response.tool is None


def test_openai_call_response_with_tools() -> None:
    """Tests the `OpenAICallResponse` class with tools."""

    class FormatBook(OpenAITool):
        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    tool_call = ChatCompletionMessageToolCall(
        id="id",
        function=Function(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
        type="function",
    )
    completion = ChatCompletion(
        id="id",
        choices=[
            Choice(
                finish_reason="stop",
                index=0,
                message=ChatCompletionMessage(
                    content="content", role="assistant", tool_calls=[tool_call]
                ),
            )
        ],
        created=0,
        model="gpt-4o",
        object="chat.completion",
    )
    call_response = OpenAICallResponse(
        metadata={},
        response=completion,
        tool_types=[FormatBook],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    tools = call_response.tools
    tool = call_response.tool
    assert tools and len(tools) == 1 and tools[0] == tool
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    output = tool.call()
    assert output == "The Name of the Wind by Patrick Rothfuss"
    assert call_response.tool_message_params([(tool, output)]) == [
        ChatCompletionToolMessageParam(
            role="tool",
            content=output,
            tool_call_id=tool_call.id,
            name="FormatBook",  # type: ignore
        )
    ]



from typing import Literal

from pydantic import BaseModel, Field

from mirascope.core import openai, prompt_template


class User(BaseModel):
    name: str
    email: str
    past_purchases: list[str]
    past_charges: list[float]
    payment_method: str
    password: str
    security_question: str
    security_answer: str


def get_user_by_email(email: str):
    if email == "johndoe@gmail.com":
        return User(
            name="John Doe",
            email="johndoe@gmail.com",
            past_purchases=["TV", "Microwave", "Chair"],
            past_charges=[349.99, 349.99, 99.99, 44.99],
            payment_method="AMEX 1234 1234 1234 1234",
            password="password1!",
            security_question="Childhood Pet Name",
            security_answer="Piddles",
        )
    else:
        return None


def get_sale_items():
    return "Sale items: we have a monitor at half off for $80!"


def get_rewards(user: User):
    if sum(user.past_charges) > 300:
        return "Rewards: for your loyalty, you get 10% off your next purchase!"
    else:
        return "Rewards: you have no rewards available right now."


def get_billing_details(user: User):
    return {
        "user_email": user.email,
        "user_name": user.name,
        "past_purchases": user.past_purchases,
        "past_charges": user.past_charges,
    }


def get_account_details(user: User):
    return {
        "user_email": user.email,
        "user_name": user.name,
        "password": user.password,
        "security_question": user.security_question,
        "security_answer": user.security_answer,
    }


def route_to_agent(
    agent_type: Literal["billing", "sale", "support"], summary: str
) -> None:
    """Routes the call to an appropriate agent with a summary of the issue."""
    print(f"Routed to: {agent_type}\nSummary:\n{summary}")


class CallClassification(BaseModel):
    calltype: Literal["billing", "sale", "support"] = Field(
        ...,
        description="""The classification of the customer's issue into one of the 3: 
        'billing' for an inquiry about charges or payment methods,
        'sale' for making a purchase,
        'support' for general FAQ or account-related questions""",
    )
    reasoning: str = Field(
        ...,
        description="""A brief description of why the customer's issue fits into the\
              chosen category""",
    )
    user_email: str = Field(..., description="email of the user in the chat")


@openai.call(model="gpt-4o-mini", response_model=CallClassification)
@prompt_template(
    """
    Classify the following transcript between a customer and the service bot:
    {transcript}
    """
)
def classify_transcript(transcript: str): ...


@openai.call(model="gpt-4o-mini", tools=[route_to_agent])
@prompt_template(
    """
    SYSTEM:
    You are an intermediary between a customer's interaction with a support chatbot and
    a real life support agent. Organize the context so that the agent can best
    facilitate the customer, but leave in details or raw data that the agent would need
    to verify a person's identity or purchase. Then, route to the appropriate agent.

    USER:
    {context}
    """
)
def handle_ticket(transcript: str) -> openai.OpenAIDynamicConfig:
    context = transcript
    call_classification = classify_transcript(transcript)
    user = get_user_by_email(call_classification.user_email)
    if isinstance(user, User):
        if call_classification.calltype == "billing":
            context += str(get_billing_details(user))
        elif call_classification.calltype == "sale":
            context += get_sale_items()
            context += get_rewards(user)
        elif call_classification.calltype == "support":
            context += str(get_account_details(user))
    else:
        context = "This person cannot be found in our system."

    return {"computed_fields": {"context": context}}


billing_transcript = """
BOT: Please enter your email.
CUSTOMER: johndoe@gmail.com
BOT: What brings you here today?
CUSTOMER: I purchased a TV a week ago but the charge is showing up twice on my bank \
statement. Can I get a refund?
"""

sale_transcript = """
BOT: Please enter your email.
CUSTOMER: johndoe@gmail.com
BOT: What brings you here today?
CUSTOMER: I'm looking to buy a new monitor. Any discounts available?
"""

support_transcript = """
BOT: Please enter your email.
CUSTOMER: johndoe@gmail.com
BOT: What brings you here today?
CUSTOMER: I forgot my site password and I'm also locked out of my email, how else can I
verify my identity?
"""

for transcript in [billing_transcript, sale_transcript, support_transcript]:
    response = handle_ticket(transcript)
    if tool := response.tool:
        tool.call()



"""usage docs: learn/calls.md#provider-specific-parameters"""

from __future__ import annotations

from anthropic.types.completion_create_params import Metadata
from anthropic.types.message_create_params import ToolChoice
from httpx import Timeout
from typing_extensions import NotRequired, Required

from ..base import BaseCallParams


class AnthropicCallParams(BaseCallParams):
    """The parameters to use when calling the Anthropic API.

    [Anthropic API Reference](https://docs.anthropic.com/en/api/messages)

    Attributes:
        max_tokens: ...
        tool_choice: ...
        metadata: ...
        stop_sequences: ...
        temperature: ...
        top_k: ...
        top_p: ...
        timeout: ...
    """

    max_tokens: Required[int]
    tool_choice: NotRequired[ToolChoice | None]
    metadata: NotRequired[Metadata | None]
    stop_sequences: NotRequired[list[str] | None]
    temperature: NotRequired[float | None]
    top_k: NotRequired[int | None]
    top_p: NotRequired[float | None]
    timeout: NotRequired[float | Timeout | None]



from pydantic import ConfigDict, Field

from mirascope.core import BaseTool


class FormatBook(BaseTool):
    """Format a book's title and author."""

    title: str = Field(..., examples=["The Name of the Wind"])
    author: str = Field(..., examples=["Rothfuss, Patrick"])

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {"title": "The Name of the Wind", "author": "Rothfuss, Patrick"}
            ]
        }
    )

    def call(self) -> str:
        return f"{self.title} by {self.author}"



from mirascope.core import openai, prompt_template


def format_book(title: str, author: str) -> str:
    """Format a book's title and author."""
    return f"{title} by {author}"


@openai.call(model="gpt-4", tools=[format_book], stream=True)
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


for chunk, tool in recommend_book("fantasy"):
    if tool:
        print(tool.call())
    else:
        print(chunk.content, end="", flush=True)



from pydantic import BaseModel

from mirascope.core import openai, prompt_template


class Book(BaseModel):
    title: str
    author: str


@openai.call(model="gpt-4o-mini", response_model=Book, stream=True)
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


book_stream = recommend_book("science fiction")
for partial_book in book_stream:
    print(partial_book)



"""Tests the `groq.call` module."""

import sys
from unittest.mock import MagicMock, patch

from mirascope.core.groq import _utils
from mirascope.core.groq.call_params import GroqCallParams
from mirascope.core.groq.call_response import GroqCallResponse
from mirascope.core.groq.call_response_chunk import GroqCallResponseChunk
from mirascope.core.groq.dynamic_config import GroqDynamicConfig
from mirascope.core.groq.stream import GroqStream
from mirascope.core.groq.tool import GroqTool


def test_groq_call() -> None:
    """Tests the `groq_call` decorator."""

    if "mirascope.core.groq._call" in sys.modules:
        del sys.modules["mirascope.core.groq._call"]

    with patch(
        "mirascope.core.base.call_factory", new_callable=MagicMock
    ) as mock_call_factory:
        import mirascope.core.groq._call  # noqa: F401

        mock_call_factory.assert_called_once_with(
            TCallResponse=GroqCallResponse,
            TCallResponseChunk=GroqCallResponseChunk,
            TDynamicConfig=GroqDynamicConfig,
            TToolType=GroqTool,
            TStream=GroqStream,
            TCallParams=GroqCallParams,
            default_call_params={},
            setup_call=_utils.setup_call,
            get_json_output=_utils.get_json_output,
            handle_stream=_utils.handle_stream,
            handle_stream_async=_utils.handle_stream_async,
        )



from openai import OpenAI

from mirascope.core import openai, prompt_template

PII_DEFINITION = """
Any representation of information that permits the identity of an individual to whom 
the information applies to be reasonably inferred by either direct or indirect means. 
Further, PII is defined as information: (i) that directly identifies an 
individual (e.g., name, address, social security number or other identifying 
number or code, telephone number, email address, etc.) or (ii) by which an agency 
intends to identify specific individuals in conjunction with other data elements, 
i.e., indirect identification. (These data elements may include a combination of gender, 
race, birth date, geographic indicator, and other descriptors). Additionally, 
information permitting the physical or online contacting of a specific individual is 
the same as personally identifiable information. This information can be maintained 
in either paper, electronic or other media.
"""

PII_ARTICLE = """
John Doe, born on 12/07/1985, resides at 123 Ruecker Harbor in Goodwinshire, WI. 
His Social Security number is 325-21-4386 and he can be reached at (123) 456-7890. 
"""


@openai.call(
    model="llama3.1",
    client=OpenAI(base_url="http://localhost:11434/v1", api_key="ollama"),
    json_mode=True,
    response_model=bool,
)
@prompt_template(
    """
    SYSTEM:
    You are an expert at identifying personally identifiable information (PII).
    Using the following definition of PII, 
    determine if the article contains PII with True or False?

    Definition of PII:
    {PII_DEFINITION}

    USER:
    {article}
    """
)
def check_if_pii_exists(article: str) -> openai.OpenAIDynamicConfig:
    return {"computed_fields": {"PII_DEFINITION": PII_DEFINITION}}


@openai.call(
    model="llama3.1",
    client=OpenAI(base_url="http://localhost:11434/v1", api_key="ollama"),
)
@prompt_template(
    """
    SYSTEM:
    You are an expert at redacting personally identifiable information (PII).
    Replace the PII in the following article with context words.

    If PII exists in the article, replace it with context words. For example, if the
    phone number is 123-456-7890, replace it with [PHONE_NUMBER].

    USER:
    {article}
    """
)
def scrub_pii(article: str): ...


def run():
    does_pii_exist = check_if_pii_exists(PII_ARTICLE)
    if does_pii_exist:
        return scrub_pii(PII_ARTICLE)
    else:
        return "No PII found in the article."


print(run())



"""Tests the `anthropic.call` module."""

import sys
from unittest.mock import MagicMock, patch

from mirascope.core.anthropic import _utils
from mirascope.core.anthropic.call_params import AnthropicCallParams
from mirascope.core.anthropic.call_response import AnthropicCallResponse
from mirascope.core.anthropic.call_response_chunk import AnthropicCallResponseChunk
from mirascope.core.anthropic.dynamic_config import AnthropicDynamicConfig
from mirascope.core.anthropic.stream import AnthropicStream
from mirascope.core.anthropic.tool import AnthropicTool


def test_anthropic_call() -> None:
    """Tests the `anthropic_call` decorator."""

    if "mirascope.core.anthropic._call" in sys.modules:
        del sys.modules["mirascope.core.anthropic._call"]

    with patch(
        "mirascope.core.base.call_factory", new_callable=MagicMock
    ) as mock_call_factory:
        import mirascope.core.anthropic._call  # noqa: F401

        mock_call_factory.assert_called_once_with(
            TCallResponse=AnthropicCallResponse,
            TCallResponseChunk=AnthropicCallResponseChunk,
            TDynamicConfig=AnthropicDynamicConfig,
            TStream=AnthropicStream,
            TToolType=AnthropicTool,
            TCallParams=AnthropicCallParams,
            default_call_params=AnthropicCallParams(max_tokens=1000),
            setup_call=_utils.setup_call,
            get_json_output=_utils.get_json_output,
            handle_stream=_utils.handle_stream,
            handle_stream_async=_utils.handle_stream_async,
        )



from mirascope.core import BaseMessageParam, BasePrompt, prompt_template


@prompt_template(
    """
    SYSTEM: You are the world's greatest librarian.
    MESSAGES: {history}
    USER: {query}
    """
)
class BookRecommendationPrompt(BasePrompt):
    history: list[BaseMessageParam]
    query: str


prompt = BookRecommendationPrompt(
    history=[
        BaseMessageParam(role="user", content="What should I read next?"),
        BaseMessageParam(
            role="assistant",
            content="I recommend 'The Name of the Wind' by Patrick Rothfuss",
        ),
    ],
    query="Anything similar you would recommend?",
)
print(prompt.message_params())
# > [
#     BaseMessageParam(role='system', content="You are the world's greatest librarian."),
#     BaseMessageParam(role='user', content='What should I read next?'),
#     BaseMessageParam(role='assistant', content="I recommend 'The Name of the Wind' by Patrick Rothfuss"),
#     BaseMessageParam(role='user', content='Anything similar you would recommend?')
#   ]



"""usage docs: learn/calls.md#provider-specific-parameters"""

from __future__ import annotations

from google.generativeai.types import GenerationConfig, RequestOptions
from google.generativeai.types.content_types import ToolConfigType
from google.generativeai.types.safety_types import SafetySettingOptions
from typing_extensions import NotRequired

from ..base import BaseCallParams


class GeminiCallParams(BaseCallParams):
    """The parameters to use when calling the Gemini API.

    [Gemini API Reference](https://ai.google.dev/gemini-api/docs/text-generation?lang=python)

    Attributes:
        generation_config: ...
        safety_settings: ...
        request_options: ...
        tool_config: ...
    """

    generation_config: NotRequired[GenerationConfig]
    safety_settings: NotRequired[SafetySettingOptions]
    request_options: NotRequired[RequestOptions]
    tool_config: NotRequired[ToolConfigType]
